@@ -13,13 +13,44 @@
 #'
 #' `compare()` is an alternative to [all.equal()].
 #'
-#' @param x,y Objects to compare. `y` is treated as the reference object
-#'   so messages describe how `x` is different to `y`
+#' @section Controlling comparisons:
+#'
+#' There are two ways for an object (rather than the person calling `compare()`
+#' or `expect_equal()` to control how it is compared to other objects.
+#' First, if the object has an S3 class, you can provide a [compare_proxy()]
+#' method that provides an alternative representation of the object; this is
+#' particularly useful if important data is stored outside of R, e.g. in
+#' an external pointer.
+#'
+#' Alternatively, you can attach an attribute called `"waldo_opts"` to your
+#' object. This should be a list of compare options, using the same names
+#' and possible values as the arguments to this function. This option
+#' is ignored by default (`ignore_attr`) so that you can set the options in
+#' the object that you control. (If you don't want to see the attributes
+#' interactively, you could attach them in a [compare_proxy()] method.)
+#'
+#' Options supplied in this way also affect all the children. This means
+#' options are applied in the following order, from lowest to highest
+#' precedence:
+#'
+#' 1.  Defaults from `compare()`.
+#' 1.  The `waldo_opts` for the parents of `x`.
+#' 1.  The `waldo_opts` for the parents of `y`.
+#' 1.  The `waldo_opts` for `x`.
+#' 1.  The `waldo_opts` for `y`.
+#' 1.  User-specified arguments to `compare()`.
+#'
+#' Use these techniques with care. If you accidentally cover up an important
+#' difference you can create a confusing situation where `x` and `y` behave
+#' differently but `compare()` reports no differences in the underlying objects.
+#'
+#' @param x,y Objects to compare. `x` is treated as the reference object
+#'   so messages describe how `y` is different to `x`.
 #' @param x_arg,y_arg Name of `x` and `y` arguments, used when generated paths
 #'   to internal components. These default to "old" and "new" since it's
 #'   most natural to supply the previous value then the new value.
 #' @param ... A handful of other arguments are supported with a warning for
-#'   backward compatability. These include:
+#'   backward comparability. These include:
 #'
 #'   * `all.equal()` arguments `checkNames` and `check.attributes`
 #'   * `testthat::compare()` argument `tol`
@@ -44,6 +75,9 @@
 #'   only its printed representation.
 #' @param ignore_attr Ignore differences in specified attributes?
 #'   Supply a character vector to ignore differences in named attributes.
+#'   By default the `"waldo_opts"` attribute is listed in `ignore_attr` so
+#'   that changes to it are not reported; if you customize `ignore_attr`, you
+#'   will probably want to do this yourself.
 #'
 #'   For backward compatibility with `all.equal()`, you can also use `TRUE`,
 #'   to all ignore differences in all attributes. This is not generally
@@ -56,6 +90,12 @@
 #' @param ignore_encoding Ignore string encoding? `TRUE` by default, because
 #'   this is R's default behaviour. Use `FALSE` when specifically concerned
 #'   with the encoding, not just the value of the string.
+#' @param list_as_map Compare lists as if they are mappings between names and
+#'   values. Concretely, this drops `NULLs` in both objects and sorts named
+#'   components.
+#' @param quote_strings Should strings be surrounded by quotes? If `FALSE`,
+#'   only side-by-side and line-by-line comparisons will be used, and there's
+#'   no way to distinguish between `NA` and `"NA"`.
 #' @returns A character vector with class "waldo_compare". If there are no
 #'   differences it will have length 0; otherwise each element contains the
 #'   description of a single difference.
@@ -82,15 +122,18 @@
 #' # Otherwise they're compared by position
 #' compare(list("x", "y"), list("x", "z"))
 #' compare(list(x = "x", x = "y"), list(x = "x", y = "z"))
+#'
 compare <- function(x, y, ...,
                     x_arg = "old", y_arg = "new",
                     tolerance = NULL,
                     max_diffs = if (in_ci()) Inf else 10,
                     ignore_srcref = TRUE,
-                    ignore_attr = FALSE,
+                    ignore_attr = "waldo_opts",
                     ignore_encoding = TRUE,
                     ignore_function_env = FALSE,
-                    ignore_formula_env = FALSE
+                    ignore_formula_env = FALSE,
+                    list_as_map = FALSE,
+                    quote_strings = TRUE
                     ) {
 
   opts <- compare_opts(
@@ -101,15 +144,36 @@ compare <- function(x, y, ...,
     ignore_attr = ignore_attr,
     ignore_encoding = ignore_encoding,
     ignore_formula_env = ignore_formula_env,
-    ignore_function_env = ignore_function_env
+    ignore_function_env = ignore_function_env,
+    list_as_map = list_as_map,
+    quote_strings = quote_strings
   )
+  # Record options overridden by user
+  opts$user_specified <- intersect(names(opts), names(match.call()))
+
   out <- compare_structure(x, y, paths = c(x_arg, y_arg), opts = opts)
   new_compare(out, max_diffs)
 }
 
-
 compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts()) {
-  if (is_reference(x, y)) {
+  if (!is_missing(x)) {
+    proxy <- compare_proxy(x, paths[[1]])
+    x <- proxy$object
+    paths[[1]] <- proxy$path
+  }
+  if (!is_missing(y)) {
+    proxy <- compare_proxy(y, paths[[2]])
+    y <- proxy$object
+    paths[[2]] <- proxy$path
+  }
+
+  opts <- merge_lists(opts,
+    attr(x, "waldo_opts"),
+    attr(y, "waldo_opts"),
+    opts[opts$user_specified]
+  )
+
+  if (is_identical(x, y, opts)) {
     return(character())
   }
 
@@ -122,8 +186,10 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
     return(term)
   }
 
-  x <- compare_proxy(x)
-  y <- compare_proxy(y)
+  if (is_list(x) && opts$list_as_map) {
+    x <- as_map(x)
+    y <- as_map(y)
+  }
 
   out <- character()
 
@@ -159,6 +225,10 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
 
   # Then contents
   if (is_list(x) || is_pairlist(x) || is.expression(x)) {
+    if (is.data.frame(x) && is.data.frame(y)) {
+      out <- c(out, compare_data_frame(x, y, paths, opts = opts))
+    }
+
     x <- unclass(x)
     y <- unclass(y)
 
@@ -168,18 +238,34 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
     } else {
       out <- c(out, compare_by_pos(x, y, paths, opts))
     }
-  } else if (is_environment(x)) {
-    if (env_has(x, ".__enclos_env__")) {
-      # enclosing env of methods is object env
-      opts$ignore_function_env <- TRUE
-      x_fields <- as.list(x, sorted = TRUE)
-      y_fields <- as.list(y, sorted = TRUE)
-      x_fields$.__enclos_env__ <- NULL
-      y_fields$.__enclos_env__ <- NULL
 
-      out <- c(out, compare_structure(x_fields, y_fields, paths, opts = opts))
-    } else {
+  } else if (is_environment(x)) {
+    if (is_seen(list(x, y), opts$seen$envs)) {
+      # Only report difference between pairs of environments once
+      return(out)
+    } else if (is_named_env(x) || is_named_env(y)) {
+      # Compare by reference
       out <- c(out, should_be("<env:{env_label(x)}>", "<env:{env_label(y)}>"))
+    } else {
+      # Compare by value
+      x_fields <- as.list.environment(x, all.names = TRUE)
+      y_fields <- as.list.environment(y, all.names = TRUE)
+      # Can't use as.list(sorted = TRUE), https://github.com/r-lib/waldo/issues/84
+      if (length(x_fields) > 0) x_fields <- x_fields[order(names(x_fields))]
+      if (length(y_fields) > 0) y_fields <- y_fields[order(names(y_fields))]
+
+      if (env_has(x, ".__enclos_env__")) {
+        # enclosing env of R6 methods is object env
+        opts$ignore_function_env <- TRUE
+        x_fields$.__enclos_env__ <- NULL
+        y_fields$.__enclos_env__ <- NULL
+      }
+
+      opts$seen$envs <- c(opts$seen$envs, list(list(x, y)))
+      out <- c(out, compare_structure(x_fields, y_fields, paths, opts = opts))
+      out <- c(out, compare_structure(
+        parent.env(x), parent.env(y), paste0("parent.env(", paths, ")"), opts = opts)
+      )
     }
   } else if (is_closure(x)) {
     if (opts$ignore_function_env) {
@@ -215,40 +301,82 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
         max_diffs = opts$max_diffs
       ))
     }
-    attributes(x) <- NULL
-    attributes(y) <- NULL
 
-    out <- c(out, switch(typeof(x),
-      integer = ,
-      complex = ,
-      double = compare_numeric(x, y, paths,
-        tolerance = opts$tolerance,
-        max_diffs = opts$max_diffs
-      ),
-      logical = compare_logical(x, y, paths, max_diffs = opts$max_diffs),
-      raw = ,
-      character = compare_character(x, y, paths, max_diffs = opts$max_diffs)
-    ))
+    out <- c(out, compare_vector(x, y, paths = paths, opts = opts))
   } else if (typeof(x) == "externalptr") {
     x <- utils::capture.output(print(x))
     y <- utils::capture.output(print(y))
     out <- c(out, should_be("{x}", "{y}"))
-  } else if (!isS4(x)) {
-    abort(glue("{paths[[1]]}: unsupported type {typeof(x)}"))
+  } else if (typeof(x) == "char") {
+    x <- paste0("CHARSXP: ", deparse(x))
+    y <- paste0("CHARSXP: ", deparse(y))
+    out <- c(out, should_be("{x}", "{y}"))
+  } else if (typeof(x) == "...") {
+    # Unevaluated dots are unlikely to lead to any significant differences
+    # in behaviour (they're usually captured incidentally) so we just
+    # ignore
+  } else if (typeof(x) != "S4") {
+    abort(glue("{paths[[1]]}: unsupported type {typeof(x)}"), call = NULL)
   }
 
   out
 }
 
+is_named_env <- function(x) {
+  environmentName(x) != ""
+}
+is_seen <- function(x, envs) {
+  for (env in envs) {
+    if (identical(x, env)) {
+      return(TRUE)
+    }
+  }
+  FALSE
+}
+
+# Fast path for "identical" elements - in the long run we'd eliminate this
+# by re-writing all of waldo in C, but this gives us a nice performance boost
+# with for a relatively low cost in the meantime.
+is_identical <- function(x, y, opts) {
+  # These comparisons aren't 100% correct because they don't affect comparison
+  # of character vectors/functions further down the tree. But I think that's
+  # unlikely to have an impact in practice since they're opt-in.
+  if (is_character(x) && is_character(y) && !opts$ignore_encoding) {
+    identical(x, y) && identical(Encoding(x), Encoding(y))
+  } else if (is_function(x) && is_function(y) && !opts$ignore_srcref) {
+    identical(x, y) && identical(attr(x, "srcref"), attr(y, "srcref"))
+  } else {
+    identical(x, y)
+  }
+}
+
 compare_terminate <- function(x, y, paths,
                               tolerance = NULL,
                               ignore_attr = FALSE) {
+  type_x <- friendly_type_of(x)
+  type_y <- friendly_type_of(y)
+  if (is_missing(x) && !is_missing(y)) {
+    type_y <- col_d(type_y)
+  } else if (!is_missing(x) && is_missing(y)) {
+    type_x <- col_a(type_x)
+  } else {
+    type_x <- col_c(type_x)
+    type_y <- col_c(type_y)
+  }
+
+  type_mismatch_msg <- should_be("{type_x}{short_val(x)}", "{type_y}{short_val(y)}")
+
+  # missing needs to be treated here because `typeof(missing_arg())` is symbol
+  if (is_missing(x) != is_missing(y)) {
+    return(type_mismatch_msg)
+  }
 
   if (typeof(x) == typeof(y) && oo_type(x) == oo_type(y)) {
     return(character())
   }
 
-  if (isTRUE(ignore_attr) && (typeof(x) == typeof(y))) {
+  ignore_class <- isTRUE(ignore_attr) || "class" %in% ignore_attr
+  if (ignore_class && (typeof(x) == typeof(y))) {
     return(character())
   }
 
@@ -261,18 +389,7 @@ compare_terminate <- function(x, y, paths,
     return(should_be("`{deparse(x)}`", "`{deparse(y)}`"))
   }
 
-  type_x <- friendly_type_of(x)
-  type_y <- friendly_type_of(y)
-  if (is_missing(x) && !is_missing(y)) {
-    type_y <- col_d(type_y)
-  } else if (!is_missing(x) && is_missing(y)) {
-    type_x <- col_a(type_x)
-  } else {
-    type_x <- col_c(type_x)
-    type_y <- col_c(type_y)
-  }
-
-  should_be("{type_x}{short_val(x)}", "{type_y}{short_val(y)}")
+  type_mismatch_msg
 }
 
 should_be <- function(x, y) {
@@ -321,6 +438,9 @@ compare_by_pos <- compare_by(index_pos, extract_pos, path_pos)
 path_line <- function(path, i) glue("lines({path}[[{i}]])")
 compare_by_line <- compare_by(index_pos, extract_pos, path_line)
 
+path_line1 <- function(path, i) glue("lines({path})")
+compare_by_line1 <- compare_by(index_pos, extract_pos, path_line1)
+
 path_attr <- function(path, i) {
   # from ?attributes, excluding row.names() because it's not a simple accessor
   funs <- c("comment", "class", "dim", "dimnames", "levels", "names", "tsp")
