@@ -19,16 +19,16 @@ diff_align <- function(diff, x, y) {
     }
 
     x_out <- c(x_out, switch(row$t,
-      a = c(col_x(x[x_i]), NA[y_i]),
-      c = col_c(x[x_i]),
-      d = col_d(x[x_i]),
-      x = col_x(x[x_i])
+      a = col_x(extract(x, c(x_i, NA[y_i]))),
+      c = col_c(extract(x, x_i)),
+      d = col_d(extract(x, x_i)),
+      x = col_x(extract(x, x_i))
     ))
     y_out <- c(y_out, switch(row$t,
-      a = col_a(y[y_i]),
-      c = col_c(y[y_i]),
-      d = c(col_x(y[y_i]), NA[x_i]),
-      x = col_x(y[y_i])
+      a = col_a(extract(y, y_i)),
+      c = col_c(extract(y, y_i)),
+      d = col_x(extract(y, c(y_i, NA[x_i]))),
+      x = col_x(extract(y, y_i))
     ))
 
     x_idx <- c(x_idx, x_i[x_i != 0], if (row$t == "a") NA[y_i])
@@ -59,6 +59,12 @@ diff_align <- function(diff, x, y) {
   )
 }
 
+extract <- function(x, idx) {
+  out <- x[idx]
+  out[is.na(idx)] <- ""
+  out
+}
+
 # Only want to show slice if it's partial
 make_slice <- function(x, idx) {
   if (all(is.na(idx))) {
@@ -73,11 +79,10 @@ make_slice <- function(x, idx) {
   }
 }
 
-col_a <- function(x) ifelse(is.na(x), NA, cli::col_blue(x))
-col_d <- function(x) ifelse(is.na(x), NA, cli::col_yellow(x))
-col_c <- function(x) ifelse(is.na(x), NA, cli::col_green(x))
-col_x <- function(x) ifelse(is.na(x), NA, cli::col_grey(x))
-
+col_a <- function(x) cli::col_blue(x)
+col_d <- function(x) cli::col_yellow(x)
+col_c <- function(x) cli::col_green(x)
+col_x <- function(x) cli::col_grey(x)
 
 # values ------------------------------------------------------------------
 
@@ -85,24 +90,30 @@ diff_element <- function(x, y, paths = c("x", "y"),
                          quote = "\"",
                          justify = "left",
                          max_diffs = 10,
-                         width = getOption("width")) {
+                         width = getOption("width"),
+                         is_string = FALSE) {
+
+  # Must quote before comparison to ensure that "NA" and NA_character
+  # have different representation
+  if (!is.null(quote)) {
+    x <- encodeString(unclass(x), quote = quote)
+    y <- encodeString(unclass(y), quote = quote)
+  }
+
   diff <- ses_shortest(x, y)
   if (length(diff) == 0) {
     return(new_compare())
   }
 
-  if (!is.null(quote)) {
-    x <- encodeString(x, quote = quote)
-    y <- encodeString(y, quote = quote)
-  }
-
   format <- lapply(diff, format_diff_matrix,
     x = x,
     y = y,
     paths = paths,
     justify = justify,
     width = width,
-    max_diffs = max_diffs
+    max_diffs = max_diffs,
+    # Paired comparisons are confusing for unquoted strings
+    use_paired = !is_string || !is.null(quote)
   )
   new_compare(unlist(format, recursive = FALSE))
 }
@@ -110,10 +121,10 @@ diff_element <- function(x, y, paths = c("x", "y"),
 format_diff_matrix <- function(diff, x, y, paths,
                                justify = "left",
                                width = getOption("width"),
-                               max_diffs = 10) {
+                               max_diffs = 10,
+                               use_paired = TRUE) {
   alignment <- diff_align(diff, x, y)
   mat <- rbind(alignment$x, alignment$y)
-  mat[is.na(mat)] <- ""
 
   n <- min(ncol(mat), max_diffs)
   n_trunc <- ncol(mat) - n
@@ -123,16 +134,18 @@ format_diff_matrix <- function(diff, x, y, paths,
   y_path_label <- label_path(paths[[2]], alignment$y_slice)
 
   # Paired lines ---------------------------------------------------------------
-  mat_out <- cbind(paste0("`", c(x_path_label, y_path_label), "`:"), mat)
-  if (n_trunc > 0) {
-    mat_out <- mat_out[, seq_len(n + 1)]
-    mat_out <- cbind(mat_out, c(paste0("and ", n_trunc, " more..."), "..."))
-  }
-  out <- apply(mat_out, 2, fansi_align, justify = justify)
-  rows <- apply(out, 1, paste, collapse = " ")
+  if (use_paired) {
+    mat_out <- cbind(paste0("`", c(x_path_label, y_path_label), "`:"), mat)
+    if (n_trunc > 0) {
+      mat_out <- mat_out[, seq_len(n + 1)]
+      mat_out <- cbind(mat_out, c(paste0("and ", n_trunc, " more..."), "..."))
+    }
+    out <- apply(mat_out, 2, fansi_align, justify = justify)
+    rows <- apply(out, 1, paste, collapse = " ")
 
-  if (fansi::nchar_ctl(rows[[1]]) <= width) {
-    return(paste0(rows, collapse = "\n"))
+    if (fansi::nchar_ctl(rows[[1]]) <= width) {
+      return(paste0(rows, collapse = "\n"))
+    }
   }
 
   # Side-by-side ---------------------------------------------------------------
@@ -172,15 +185,42 @@ format_diff_matrix <- function(diff, x, y, paths,
 
 line_by_line <- function(x, y, diff, max_diffs = 10) {
   lines <- character()
+  if (nrow(diff) == 0) {
+    return(lines)
+  }
 
   line_a <- function(x) if (length(x) > 0) col_a(paste0("+ ", names(x), x))
   line_d <- function(x) if (length(x) > 0) col_d(paste0("- ", names(x), x))
   line_x <- function(x) if (length(x) > 0) col_x(paste0("  ", names(x), x))
 
-  n <- min(max_diffs, nrow(diff))
-  n_trunc <- nrow(diff) - n
+  diff_lengths <- cumsum(pmax(diff$x2 - diff$x1, diff$y2 - diff$y1) + 1)
+  all_diff_lengths <- last(diff_lengths)
+  if (all_diff_lengths > max_diffs) {
+    diffs_ok <- which(stats::lag(diff_lengths, 0) <= max_diffs)
 
-  for (i in seq_len(n)) {
+    if (length(diffs_ok) == 0) {
+      diff_ok <- 0
+      diff_length_partial <- max_diffs
+    } else {
+      diff_ok <- last(diffs_ok)
+      diff_length_partial <- max_diffs - diff_lengths[[diff_ok]]
+    }
+
+    if (diff_length_partial > 0) {
+      partial_diff <- diff[diff_ok + 1, ]
+      partial_diff$x2 <- min(partial_diff$x2, partial_diff$x1 + diff_length_partial - 1)
+      partial_diff$y2 <- min(partial_diff$y2, partial_diff$y1 + diff_length_partial - 1)
+    } else {
+      partial_diff <- NULL
+    }
+    diff <- rbind(diff[seq_len(diff_ok), ], partial_diff)
+
+    n_trunc <- all_diff_lengths - max_diffs
+  } else {
+    n_trunc <- 0
+  }
+
+  for (i in seq_len(nrow(diff))) {
     row <- diff[i, , drop = FALSE]
     x_i <- seq2(row$x1, row$x2)
     y_i <- seq2(row$y1, row$y2)
@@ -220,3 +260,7 @@ label_path <- function(path, slice) {
 label_idx <- function(idx) {
   ifelse(is.na(idx), "", paste0("[", idx, "]"))
 }
+
+last <- function(x) {
+  x[[length(x)]]
+}
