@@ -6,6 +6,9 @@ num_equal <- function(x, y, tolerance = .Machine$double.eps ^ 0.5) {
   if (any(is.na(x) != is.na(y))) {
     return(FALSE)
   }
+  if (any(is.nan(x) != is.nan(y))) {
+    return(FALSE)
+  }
 
   attributes(x) <- NULL
   attributes(y) <- NULL
@@ -23,8 +26,8 @@ num_equal <- function(x, y, tolerance = .Machine$double.eps ^ 0.5) {
   avg_diff <- mean(abs(x_diff - y_diff))
   avg_y <- mean(abs(y_diff))
 
-  # compute relative difference when y is "large"
-  if (avg_y > tolerance) {
+  # compute relative difference when y is "large" but finite
+  if (is.finite(avg_y) && avg_y > tolerance) {
     avg_diff <- avg_diff / avg_y
   }
 
