@@ -23,6 +23,10 @@ test_that("character comparison", {
   })
 })
 
+test_that("NA and 'NA' compare differently", {
+  expect_snapshot(compare(NA_character_, "NA"))
+})
+
 test_that("multiline comparison", {
   expect_snapshot({
     compare_character("A\nthe apple is red\nC\n", "A\nthe apple was red\nC\n")
@@ -85,6 +89,18 @@ test_that("numeric comparison", {
   })
 })
 
+test_that("NAs are shown regardless of position", {
+  expect_snapshot({
+    compare(c(NA, 1, 2), c(1, 2))
+    compare(c(1, NA, 2), c(1, 2))
+    compare(c(1, 2, NA), c(1, 2))
+  })
+})
+
+test_that("informative difference between NA and NaN", {
+  expect_snapshot(compare_numeric(NA_real_, NaN))
+})
+
 test_that("numeric comparison works on factors", {
   expect_snapshot({
     f1 <- factor(c("a", "b", "c"))
@@ -104,6 +120,14 @@ test_that("shows row-by-row diff for numeric matrices", {
   })
 })
 
+test_that("but not for arrays", {
+  expect_snapshot({
+    x <- y <- array(1:4, c(1, 2, 2))
+    y[1, 2, 2] <- 5L
+    compare(x, y)
+  })
+})
+
 test_that("falls back to regular display if printed representation the same", {
   expect_snapshot({
     x <- y <- matrix(1:4, nrow = 2)
@@ -136,6 +160,13 @@ test_that("don't use format if numeric & within tolerance", {
   })
 })
 
+test_that("can compare complex numbers", {
+  expect_snapshot({
+    compare(1:2 + 1i, 2 + 1i)
+    compare(1:2 + 1i, 1:2 + 2i)
+  })
+})
+
 test_that("logical comparisons minimise extraneous diffs", {
   x1 <- x2 <- rep(TRUE, 50)
   x2[c(1, 25, 50)] <- FALSE
